@@ -143,29 +143,19 @@ static void
 gupnp_control_point_dispose (GObject *object)
 {
         GUPnPControlPoint *control_point;
+        GSSDPResourceBrowser *browser;
         GObjectClass *object_class;
 
         control_point = GUPNP_CONTROL_POINT (object);
+        browser = GSSDP_RESOURCE_BROWSER (control_point);
+
+        gssdp_resource_browser_set_active (browser, FALSE);
 
         if (control_point->priv->factory) {
                 g_object_unref (control_point->priv->factory);
                 control_point->priv->factory = NULL;
         }
 
-        while (control_point->priv->devices) {
-                g_object_unref (control_point->priv->devices->data);
-                control_point->priv->devices =
-                        g_list_delete_link (control_point->priv->devices,
-                                            control_point->priv->devices);
-        }
-
-        while (control_point->priv->services) {
-                g_object_unref (control_point->priv->services->data);
-                control_point->priv->services =
-                        g_list_delete_link (control_point->priv->services,
-                                            control_point->priv->services);
-        }
-
         /* Cancel any pending description file GETs */
         while (control_point->priv->pending_gets) {
                 GetDescriptionURLData *data;
@@ -209,11 +199,135 @@ gupnp_control_point_finalize (GObject *object)
         object_class->finalize (object);
 }
 
+static GList *
+find_service_node (GUPnPControlPoint *control_point,
+                   const char        *udn,
+                   const char        *service_type)
+{
+        GList *l;
+
+        l = control_point->priv->services;
+
+        while (l) {
+                GUPnPServiceInfo *info;
+
+                info = GUPNP_SERVICE_INFO (l->data);
+
+                if ((strcmp (gupnp_service_info_get_udn (info), udn) == 0) ||
+                    (strcmp (gupnp_service_info_get_service_type (info),
+                             service_type) == 0))
+                        break;
+
+                l = l->next;
+        }
+
+        return l;
+}
+
+static GList *
+find_device_node (GUPnPControlPoint *control_point,
+                  const char        *udn)
+{
+        GList *l;
+
+        l = control_point->priv->devices;
+
+        while (l) {
+                GUPnPDeviceInfo *info;
+
+                info = GUPNP_DEVICE_INFO (l->data);
+
+                if (strcmp (udn, gupnp_device_info_get_udn (info)) == 0)
+                        break;
+
+                l = l->next;
+        }
+
+        return l;
+}
+
+static void
+create_and_report_service_proxy (GUPnPControlPoint  *control_point,
+                                 GUPnPXMLDoc        *doc,
+                                 xmlNode            *element,
+                                 const char         *udn,
+                                 const char         *service_type,
+                                 const char         *description_url,
+                                 SoupURI            *url_base)
+{
+        GUPnPServiceProxy *proxy;
+        GUPnPResourceFactory *factory;
+        GUPnPContext *context;
+
+        if (find_service_node (control_point, udn, service_type) != NULL)
+                /* We already have a proxy for this service */
+                return;
+
+        factory = gupnp_control_point_get_resource_factory (control_point);
+        context = gupnp_control_point_get_context (control_point);
+
+        /* Create proxy */
+        proxy = gupnp_resource_factory_create_service_proxy (factory,
+                                                             context,
+                                                             doc,
+                                                             element,
+                                                             udn,
+                                                             service_type,
+                                                             description_url,
+                                                             url_base);
+
+        control_point->priv->services =
+                                g_list_prepend (control_point->priv->services,
+                                                proxy);
+
+        g_signal_emit (control_point,
+                       signals[SERVICE_PROXY_AVAILABLE],
+                       0,
+                       proxy);
+}
+
+static void
+create_and_report_device_proxy (GUPnPControlPoint  *control_point,
+                                GUPnPXMLDoc        *doc,
+                                xmlNode            *element,
+                                const char         *udn,
+                                const char         *description_url,
+                                SoupURI            *url_base)
+{
+        GUPnPDeviceProxy *proxy;
+        GUPnPResourceFactory *factory;
+        GUPnPContext *context;
+
+        if (find_device_node (control_point, udn) != NULL)
+                /* We already have a proxy for this device */
+                return;
+
+        factory = gupnp_control_point_get_resource_factory (control_point);
+        context = gupnp_control_point_get_context (control_point);
+
+        proxy = gupnp_resource_factory_create_device_proxy (factory,
+                                                            context,
+                                                            doc,
+                                                            element,
+                                                            udn,
+                                                            description_url,
+                                                            url_base);
+
+        control_point->priv->devices = g_list_prepend
+                                                (control_point->priv->devices,
+                                                 proxy);
+
+        g_signal_emit (control_point,
+                       signals[DEVICE_PROXY_AVAILABLE],
+                       0,
+                       proxy);
+}
+
 /* Search @element for matching services */
 static void
 process_service_list (xmlNode           *element,
                       GUPnPControlPoint *control_point,
-                      XmlDocWrapper     *doc,
+                      GUPnPXMLDoc       *doc,
                       const char        *udn,
                       const char        *service_type,
                       const char        *description_url,
@@ -224,8 +338,6 @@ process_service_list (xmlNode           *element,
         for (element = element->children; element; element = element->next) {
                 xmlChar *prop;
                 gboolean match;
-                GUPnPContext *context;
-                GUPnPServiceProxy *proxy;
 
                 if (strcmp ((char *) element->name, "service") != 0)
                         continue;
@@ -245,28 +357,13 @@ process_service_list (xmlNode           *element,
 
                 /* Match */
 
-                /* Get context */
-                context = gupnp_control_point_get_context (control_point);
-
-                /* Create proxy */
-                proxy = gupnp_resource_factory_create_service_proxy
-                                (gupnp_control_point_get_resource_factory (control_point),
-                                 context,
-                                 doc,
-                                 element,
-                                 udn,
-                                 service_type,
-                                 description_url,
-                                 url_base);
-
-                control_point->priv->services =
-                        g_list_prepend (control_point->priv->services,
-                                        proxy);
-
-                g_signal_emit (control_point,
-                               signals[SERVICE_PROXY_AVAILABLE],
-                               0,
-                               proxy);
+                create_and_report_service_proxy (control_point,
+                                                 doc,
+                                                 element,
+                                                 udn,
+                                                 service_type,
+                                                 description_url,
+                                                 url_base);
         }
 
         g_object_unref (control_point);
@@ -276,7 +373,7 @@ process_service_list (xmlNode           *element,
 static void
 process_device_list (xmlNode           *element,
                      GUPnPControlPoint *control_point,
-                     XmlDocWrapper     *doc,
+                     GUPnPXMLDoc       *doc,
                      const char        *udn,
                      const char        *service_type,
                      const char        *description_url,
@@ -288,7 +385,6 @@ process_device_list (xmlNode           *element,
                 xmlNode *children;
                 xmlChar *prop;
                 gboolean match;
-                GUPnPContext *context;
 
                 if (strcmp ((char *) element->name, "device") != 0)
                         continue;
@@ -322,9 +418,6 @@ process_device_list (xmlNode           *element,
 
                 /* Match */
 
-                /* Get context */
-                context = gupnp_control_point_get_context (control_point);
-
                 if (service_type) {
                         /* Dive into serviceList */
                         children = xml_util_get_element (element,
@@ -340,29 +433,13 @@ process_device_list (xmlNode           *element,
                                                       description_url,
                                                       url_base);
                         }
-                } else {
-                        /* Create device proxy */
-                        GUPnPDeviceProxy *proxy;
-
-                        proxy = gupnp_resource_factory_create_device_proxy
-                                        (gupnp_control_point_get_resource_factory (control_point),
-                                         context,
-                                         doc,
-                                         element,
-                                         udn,
-                                         description_url,
-                                         url_base);
-
-                        control_point->priv->devices =
-                                g_list_prepend
-                                        (control_point->priv->devices,
-                                         proxy);
-
-                        g_signal_emit (control_point,
-                                       signals[DEVICE_PROXY_AVAILABLE],
-                                       0,
-                                       proxy);
-                }
+                } else
+                        create_and_report_device_proxy (control_point,
+                                                        doc,
+                                                        element,
+                                                        udn,
+                                                        description_url,
+                                                        url_base);
         }
 
         g_object_unref (control_point);
@@ -373,7 +450,7 @@ process_device_list (xmlNode           *element,
  */
 static void
 description_loaded (GUPnPControlPoint *control_point,
-                    XmlDocWrapper     *doc,
+                    GUPnPXMLDoc       *doc,
                     const char        *udn,
                     const char        *service_type,
                     const char        *description_url)
@@ -385,6 +462,17 @@ description_loaded (GUPnPControlPoint *control_point,
         element = xml_util_get_element ((xmlNode *) doc->doc,
                                         "root",
                                         NULL);
+        if (element == NULL) {
+                g_warning ("No 'root' element found in description document"
+                           " '%s'. Ignoring device '%s'\n",
+                           description_url,
+                           udn);
+
+                return;
+        }
+
+        if (element == NULL)
+                return;
 
         url_base = xml_util_get_child_element_content_uri (element,
                                                            "URLBase",
@@ -413,7 +501,7 @@ got_description_url (SoupSession           *session,
                      SoupMessage           *msg,
                      GetDescriptionURLData *data)
 {
-        XmlDocWrapper *doc;
+        GUPnPXMLDoc *doc;
 
         if (msg->status_code == SOUP_STATUS_CANCELLED)
                 return;
@@ -443,7 +531,7 @@ got_description_url (SoupSession           *session,
                 xml_doc = xmlRecoverMemory (msg->response_body->data,
                                             msg->response_body->length);
                 if (xml_doc) {
-                        doc = xml_doc_wrapper_new (xml_doc);
+                        doc = gupnp_xml_doc_new (xml_doc);
 
                         description_loaded (data->control_point,
                                             doc,
@@ -464,12 +552,13 @@ got_description_url (SoupSession           *session,
                                            data->control_point);
 
                         /* If no proxy was created, make sure doc is freed. */
-                        g_object_ref_sink (doc);
                         g_object_unref (doc);
                 } else
                         g_warning ("Failed to parse %s", data->description_url);
         } else
-                g_warning ("Failed to GET %s", data->description_url);
+                g_warning ("Failed to GET %s: %s",
+                           data->description_url,
+                           msg->reason_phrase);
 
         get_description_url_data_free (data);
 }
@@ -488,7 +577,7 @@ load_description (GUPnPControlPoint *control_point,
                   const char        *udn,
                   const char        *service_type)
 {
-        XmlDocWrapper *doc;
+        GUPnPXMLDoc *doc;
 
         doc = g_hash_table_lookup (control_point->priv->doc_cache,
                                    description_url);
@@ -522,7 +611,6 @@ load_description (GUPnPControlPoint *control_point,
                         return;
                 }
 
-                http_request_set_user_agent (data->message);
                 http_request_set_accept_language (data->message);
 
                 data->control_point   = control_point;
@@ -671,26 +759,13 @@ gupnp_control_point_resource_unavailable
 
         /* Find proxy */
         if (service_type) {
-                l = control_point->priv->services;
+                l = find_service_node (control_point, udn, service_type);
 
-                while (l) {
-                        GUPnPServiceInfo *info;
+                if (l) {
                         GUPnPServiceProxy *proxy;
 
-                        info = GUPNP_SERVICE_INFO (l->data);
-
-                        if ((strcmp (udn,
-                                     gupnp_service_info_get_udn (info)) != 0) ||
-                            (strcmp (service_type,
-                                     gupnp_service_info_get_service_type (info))
-                                     != 0)) {
-                                l = l->next;
-
-                                continue;
-                        }
-
                         /* Remove proxy */
-                        proxy = GUPNP_SERVICE_PROXY (info);
+                        proxy = GUPNP_SERVICE_PROXY (l->data);
 
                         cur_l = l;
                         l = l->next;
@@ -707,23 +782,13 @@ gupnp_control_point_resource_unavailable
                         g_object_unref (proxy);
                 }
         } else {
-                l = control_point->priv->devices;
+                l = find_device_node (control_point, udn);
 
-                while (l) {
-                        GUPnPDeviceInfo *info;
+                if (l) {
                         GUPnPDeviceProxy *proxy;
 
-                        info = GUPNP_DEVICE_INFO (l->data);
-
-                        if (strcmp (udn,
-                                    gupnp_device_info_get_udn (info)) != 0) {
-                                l = l->next;
-
-                                continue;
-                        }
-
                         /* Remove proxy */
-                        proxy = GUPNP_DEVICE_PROXY (info);
+                        proxy = GUPNP_DEVICE_PROXY (l->data);
 
                         cur_l = l;
                         l = l->next;
@@ -810,7 +875,7 @@ gupnp_control_point_class_init (GUPnPControlPointClass *klass)
         g_type_class_add_private (klass, sizeof (GUPnPControlPointPrivate));
 
         /**
-         * GUPnPControlPoint:resource-factory
+         * GUPnPControlPoint:resource-factory:
          *
          * The resource factory to use. Set to NULL for default factory.
          **/
@@ -828,7 +893,7 @@ gupnp_control_point_class_init (GUPnPControlPointClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPControlPoint::device-proxy-available
+         * GUPnPControlPoint::device-proxy-available:
          * @control_point: The #GUPnPControlPoint that received the signal
          * @proxy: The now available #GUPnPDeviceProxy
          *
@@ -849,7 +914,7 @@ gupnp_control_point_class_init (GUPnPControlPointClass *klass)
                               GUPNP_TYPE_DEVICE_PROXY);
 
         /**
-         * GUPnPControlPoint::device-proxy-unavailable
+         * GUPnPControlPoint::device-proxy-unavailable:
          * @control_point: The #GUPnPControlPoint that received the signal
          * @proxy: The now unavailable #GUPnPDeviceProxy
          *
@@ -870,7 +935,7 @@ gupnp_control_point_class_init (GUPnPControlPointClass *klass)
                               GUPNP_TYPE_DEVICE_PROXY);
 
         /**
-         * GUPnPControlPoint::service-proxy-available
+         * GUPnPControlPoint::service-proxy-available:
          * @control_point: The #GUPnPControlPoint that received the signal
          * @proxy: The now available #GUPnPServiceProxy
          *
@@ -891,7 +956,7 @@ gupnp_control_point_class_init (GUPnPControlPointClass *klass)
                               GUPNP_TYPE_SERVICE_PROXY);
 
         /**
-         * GUPnPControlPoint::service-proxy-unavailable
+         * GUPnPControlPoint::service-proxy-unavailable:
          * @control_point: The #GUPnPControlPoint that received the signal
          * @proxy: The now unavailable #GUPnPServiceProxy
          *
@@ -913,7 +978,7 @@ gupnp_control_point_class_init (GUPnPControlPointClass *klass)
 }
 
 /**
- * gupnp_control_point_new
+ * gupnp_control_point_new:
  * @context: A #GUPnPContext
  * @target: The search target
  *
@@ -939,7 +1004,7 @@ gupnp_control_point_new (GUPnPContext *context,
 }
 
 /**
- * gupnp_control_point_new_full
+ * gupnp_control_point_new_full:
  * @context: A #GUPnPContext
  * @factory: A #GUPnPResourceFactory
  * @target: The search target
@@ -971,7 +1036,7 @@ gupnp_control_point_new_full (GUPnPContext         *context,
 }
 
 /**
- * gupnp_control_point_get_context
+ * gupnp_control_point_get_context:
  * @control_point: A #GUPnPControlPoint
  *
  * Get the #GUPnPControlPoint associated with @control_point.
@@ -992,13 +1057,14 @@ gupnp_control_point_get_context (GUPnPControlPoint *control_point)
 }
 
 /**
- * gupnp_control_point_list_device_proxies
+ * gupnp_control_point_list_device_proxies:
  * @control_point: A #GUPnPControlPoint
  *
  * Get the #GList of discovered #GUPnPDeviceProxy objects. Do not free the list
  * nor its elements.
  *
- * Return value: a #GList of #GUPnPDeviceProxy objects.
+ * Return value: (element-type GUPnP.DeviceProxy) (transfer none):  a #GList of
+ * #GUPnPDeviceProxy objects.
  **/
 const GList *
 gupnp_control_point_list_device_proxies (GUPnPControlPoint *control_point)
@@ -1009,13 +1075,14 @@ gupnp_control_point_list_device_proxies (GUPnPControlPoint *control_point)
 }
 
 /**
- * gupnp_control_point_list_service_proxies
+ * gupnp_control_point_list_service_proxies:
  * @control_point: A #GUPnPControlPoint
  *
- * Get the #GList of discovered #GUPnPServiceProxy objects. Do not free the list
- * nor its elements.
+ * Get the #GList of discovered #GUPnPServiceProxy objects. Do not free the
+ * list nor its elements.
  *
- * Return value: a #GList of #GUPnPServiceProxy objects.
+ * Return value: (element-type: GUPnP.ServiceProxy) (transfer none): a #GList
+ * of #GUPnPServiceProxy objects.
  **/
 const GList *
 gupnp_control_point_list_service_proxies (GUPnPControlPoint *control_point)
@@ -1026,7 +1093,7 @@ gupnp_control_point_list_service_proxies (GUPnPControlPoint *control_point)
 }
 
 /**
- * gupnp_control_point_get_resource_factory
+ * gupnp_control_point_get_resource_factory:
  * @control_point: A #GUPnPControlPoint
  *
  * Get the #GUPnPResourceFactory used by the @control_point.
